﻿Imports System
Imports System.Collections.Generic
Imports System.IO
Imports System.Linq
Imports System.Windows
Imports SterlingRecipes.Models
Imports Wintellect.Sterling
Imports Wintellect.Sterling.Database

''' <summary>
'''     Set up for the Sterling database
''' </summary>
Public Class RecipeDatabase
    Inherits BaseDatabaseInstance
    ''' <summary>
    '''     Data base name
    ''' </summary>
    Public Const STERLING_RECIPES As String = "Sterling Recipes"

    ''' <summary>
    '''     Index on food names
    ''' </summary>
    Public Const IDX_FOOD_NAME As String = "IDX_FoodModel_Name"

    ''' <summary>
    '''     Index on recipe names
    ''' </summary>
    Public Const IDX_RECIPE_CATEGORYID_NAME As String = "IDX_RecipeModel_CategoryId_RecipeName"

    ''' <summary>
    '''     Sets up the tables and corresponding keys
    ''' </summary>
    ''' <returns></returns>
    Protected Overrides Function _RegisterTables() As List(Of ITableDefinition)
        Return New List(Of ITableDefinition) From {
            CreateTableDefinition(Of TombstoneModel, Type)(Function(c) c.SyncType),
            CreateTableDefinition(Of IngredientModel, Integer)(Function(i) i.Id),
            CreateTableDefinition(Of CategoryModel, Integer)(Function(c) c.Id),
            CreateTableDefinition(Of FoodModel, Integer)( _
                Function(f) f.Id).WithIndex(Of FoodModel, String, Integer)( _
                IDX_FOOD_NAME, Function(f) f.FoodName),
            CreateTableDefinition(Of MeasureModel, Integer)(Function(m) m.Id),
            CreateTableDefinition(Of RecipeModel, Integer)( _
               Function(r) r.Id).WithIndex(Of RecipeModel, Integer, String, Integer)( _
              IDX_RECIPE_CATEGORYID_NAME, Function(r) Tuple.Create(r.Category.Id, r.Name))}
        ' tombstone helper
        ' ingredient on a recipe
        ' category 
        ' food item 
        ' measurement 
        ' recipe
    End Function

    ''' <summary>
    '''     Database name
    ''' </summary>
    Public Overrides ReadOnly Property Name() As String
        Get
            Return STERLING_RECIPES
        End Get
    End Property

    ''' <summary>
    '''     Check to see if the database has been populated. If not, populate it.
    ''' </summary>
    ''' <param name="database">The database instance</param>
    Public Shared Sub CheckAndCreate(ByVal database As ISterlingDatabaseInstance)
        Const FILE_MEASURES As String = "/SterlingRecipes.Database;component/Measures.txt"
        Const FILE_CATEGORIES As String = "/SterlingRecipes.Database;component/Categories.txt"
        Const FILE_FOOD As String = "/SterlingRecipes.Database;component/Food.txt"

        ' register the triggers
        database.RegisterTrigger(New IdentityTrigger(Of FoodModel)(database))
        database.RegisterTrigger(New IdentityTrigger(Of RecipeModel)(database))
        database.RegisterTrigger(New IdentityTrigger(Of IngredientModel)(database))

        ' Categories are short so we'll do them last and check them first
        ' if any are here we've already set things up because users can't delete these
        If database.Query(Of CategoryModel, Integer)().Any() Then
            Return
        End If

        ' get rid of old data
        database.Truncate(GetType(MeasureModel))
        database.Truncate(GetType(FoodModel))

        Dim idx = 0

        For Each measure In _ParseFromResource(FILE_MEASURES, Function(line) New MeasureModel With {.Id = idx + 1, .Abbreviation = line(0), .FullMeasure = line(1)})
            measure.Id = idx + 1
            database.Save(measure)
            idx = idx + 1
        Next measure

        ' sample foods auto-generate the id
        For Each food In _ParseFromResource(FILE_FOOD, Function(line) New FoodModel With {.FoodName = line(0)}).Where(Function(_food) (Not String.IsNullOrEmpty(_food.FoodName)))
            database.Save(food)
        Next food

        Dim idx1 = 0

        For Each category In _ParseFromResource(FILE_CATEGORIES, Function(line) New CategoryModel With {.Id = idx1 + 1, .CategoryName = line(0)})
            category.Id = idx1 + 1
            database.Save(category)
            idx1 = idx1 + 1
        Next

        ' at this point the user is good, but let's give tem a sample recipe as well
        ' create a sample recipe for "AB&J Sushi Rolls" used with permission from http://www.lizziemariecuisine.com          
        Dim recipe = New RecipeModel With {.Category = database.Load(Of CategoryModel)(2),
                                           .Name = "AB&J Sushi Roll",
                                           .Instructions = "Remove crust and flatten bread with rolling pin by rolling bread on cutting board. Spread a thin layer of almond butter and jam onto bread. Roll up slice of bread to look like a sushi roll. Cut bread into four bite-size pieces using a serrated knife.",
                                           .Ingredients = New List(Of IngredientModel)()}

        ' save to get a key
        database.Save(recipe)

        Dim slice = database.Load(Of MeasureModel)(16)
        Dim tbsp = database.Load(Of MeasureModel)(12)

        Dim ingredient1 = New IngredientModel With {.Recipe = recipe, .Food = New FoodModel With {.FoodName = "Whole Wheat Bread"}, .Amount = New MeasureAmountModel With {.Units = 1.0, .Measure = slice}, .Order = 1}

        ' get an id
        database.Save(ingredient1)

        ' add ingredients
        recipe.Ingredients.Add(ingredient1)

        Dim ingredient2 = New IngredientModel With {.Recipe = recipe, .Food = New FoodModel With {.FoodName = "All-Natural Almond Butter"}, .Amount = New MeasureAmountModel With {.Units = 1.0, .Measure = tbsp}, .Order = 2}

        ' get an id
        database.Save(ingredient2)

        recipe.Ingredients.Add(ingredient2)

        Dim ingredient3 = New IngredientModel With {.Recipe = recipe, .Food = New FoodModel With {.FoodName = "All-Natural Blackberry Jam"}, .Amount = New MeasureAmountModel With {.Units = 1.0, .Measure = tbsp}, .Order = 3}

        ' get an id
        database.Save(ingredient3)
        recipe.Ingredients.Add(ingredient3)
        database.Save(recipe)

        ' get indexes written to disk
        database.Flush()
    End Sub

    ''' <summary>
    '''     Signature for a parser
    ''' </summary>
    ''' <typeparam name="T">Type to parse</typeparam>
    ''' <param name="parts">Split line</param>
    ''' <returns>Constructed entity</returns>
    ''' ReSharper disable TypeParameterCanBeVariant
    Private Delegate Function Parser(Of T)(ByVal parts() As String) As T
    ' ReSharper restore TypeParameterCanBeVariant

    ''' <summary>
    '''     Parse a list from the resource
    ''' </summary>
    ''' <typeparam name="T">The type to parse</typeparam>
    ''' <param name="resourceName">The embedded resource</param>
    ''' <param name="_parser">The parser strategy</param>
    ''' <returns>The list</returns>
    Private Shared Function _ParseFromResource(Of T)(ByVal resourceName As String, ByVal _parser As Parser(Of T)) As IEnumerable(Of T)

        Using stream = Application.GetResourceStream(New Uri(resourceName, UriKind.Relative)).Stream
            Using reader = New StreamReader(stream)
                Dim returnVal As New List(Of T)()
                Do While Not reader.EndOfStream
                    Dim line = reader.ReadLine()
                    If String.IsNullOrEmpty(line) Then
                        Exit Do
                    End If
                    returnVal.Add(_parser(line.Split(",")))
                Loop
                Return returnVal
            End Using
        End Using
    End Function
End Class