﻿Imports System.Collections.Generic
Imports System.Collections.ObjectModel
Imports System.Linq
Imports SterlingRecipes.Contracts
Imports SterlingRecipes.Database
Imports SterlingRecipes.Models
Imports UltraLight.MVVM
Imports UltraLight.MVVM.Contracts

Namespace ViewModels
    ''' <summary>
    '''     The main view model
    ''' </summary>
    Public Class MainViewModel
        Inherits BaseViewModel
        Implements IMainViewModel, ITombstoneFriendly
        ''' <summary>
        '''     Set it all up and grab the list of categories 
        ''' </summary>
        Public Sub New()
            Categories = New ObservableCollection(Of CategoryModel)()

            For Each category In App.Database.Query(Of CategoryModel, Integer)()
                Categories.Add(category.LazyValue.Value)
            Next category

            EditRecipe = New ActionCommand(Of Object)(Sub(o) _EditRecipe(), Function(o) _CanEditRecipe())

            AddRecipe = New ActionCommand(Of Object)(Sub(o) _AddRecipe())

            DeleteRecipe = New ActionCommand(Of Object)(Sub(o) _DeleteRecipe(), Function(o) _CanEditRecipe())
        End Sub

        ''' <summary>
        '''     Recipe list is 
        ''' </summary>
        Public ReadOnly Property Recipes() As IEnumerable(Of RecipeModel) Implements IMainViewModel.Recipes
            Get
                Return From r In App.Database.Query(Of RecipeModel, Integer, String, Integer)(RecipeDatabase.IDX_RECIPE_CATEGORYID_NAME)
                       Where r.Index.Item1.Equals(If(CurrentCategory Is Nothing, 0, CurrentCategory.Id))
                       Order By r.Index.Item2
                       Select New RecipeModel With {.Id = r.Key, .Name = r.Index.Item2}
            End Get
        End Property

        ''' <summary>
        '''     Edit the recipe
        ''' </summary>
        Private privateEditRecipe As IActionCommand(Of Object)
        Public Property EditRecipe() As IActionCommand(Of Object)
            Get
                Return privateEditRecipe
            End Get
            Private Set(ByVal value As IActionCommand(Of Object))
                privateEditRecipe = value
            End Set
        End Property

        Public ReadOnly Property EditRecipeAdapter() As IActionCommand(Of Object) Implements IMainViewModel.EditRecipe
            Get
                Return privateEditRecipe
            End Get
        End Property

        ''' <summary>
        '''     Add a new recipe
        ''' </summary>
        Private privateAddRecipe As IActionCommand(Of Object)
        Public Property AddRecipe() As IActionCommand(Of Object)
            Get
                Return privateAddRecipe
            End Get
            Private Set(ByVal value As IActionCommand(Of Object))
                privateAddRecipe = value
            End Set
        End Property

        Public ReadOnly Property AddRecipeAdapter() As IActionCommand(Of Object) Implements IMainViewModel.AddRecipe
            Get
                Return privateAddRecipe
            End Get
        End Property

        ''' <summary>
        '''     Delete the recipe
        ''' </summary>
        Private privateDeleteRecipe As IActionCommand(Of Object)
        Public Property DeleteRecipe() As IActionCommand(Of Object)
            Get
                Return privateDeleteRecipe
            End Get
            Private Set(ByVal value As IActionCommand(Of Object))
                privateDeleteRecipe = value
            End Set
        End Property

        Public ReadOnly Property DeleteRecipeAdapter() As IActionCommand(Of Object) Implements IMainViewModel.DeleteRecipe
            Get
                Return privateDeleteRecipe
            End Get
        End Property

        ''' <summary>
        ''' A collection for ItemViewModel objects.
        ''' </summary>
        Private privateCategories As ObservableCollection(Of CategoryModel)
        Public Property Categories() As ObservableCollection(Of CategoryModel)
            Get
                Return privateCategories
            End Get
            Private Set(ByVal value As ObservableCollection(Of CategoryModel))
                privateCategories = value
            End Set
        End Property

        Public ReadOnly Property CategoriesAdapter() As ObservableCollection(Of CategoryModel) Implements IMainViewModel.Categories
            Get
                Return privateCategories
            End Get
        End Property

        ''' <summary>
        '''     Current recipe being worked on
        ''' </summary>
        Private _currentRecipe As RecipeModel

        Public Property CurrentRecipe() As RecipeModel
            Get
                Return _currentRecipe
            End Get
            Set(ByVal value As RecipeModel)
                _currentRecipe = value
                RaisePropertyChanged(Function() CurrentRecipe)
                EditRecipe.RaiseCanExecuteChanged()
                DeleteRecipe.RaiseCanExecuteChanged()
            End Set
        End Property

        Public ReadOnly Property CurrentRecipeAdapter() As RecipeModel Implements IMainViewModel.CurrentRecipe
            Get
                Return _currentRecipe
            End Get
        End Property

        ''' <summary>
        '''     Current category selected
        ''' </summary>
        Private _currentCategory As CategoryModel

        Public Property CurrentCategory() As CategoryModel
            Get
                Return _currentCategory
            End Get
            Set(ByVal value As CategoryModel)
                _currentCategory = value
                RaisePropertyChanged(Function() CurrentCategory)
                RaisePropertyChanged(Function() Recipes)
            End Set
        End Property

        Public ReadOnly Property CurrentCategoryAdapter() As CategoryModel Implements IMainViewModel.CurrentCategory
            Get
                Return _currentCategory
            End Get
        End Property

        ''' <summary>
        '''     Add a new recipe
        ''' </summary>
        Private Sub _AddRecipe()
            UltraLightLocator.GetViewModel(Of IRecipeViewModel)().AddRecipe()
            RequestNavigation()("/Views/RecipeEdit.xaml")
        End Sub

        ''' <summary>
        '''     Edit recipe
        ''' </summary>
        Private Sub _EditRecipe()
            UltraLightLocator.GetViewModel(Of IRecipeViewModel)().EditRecipe(App.Database.Load(Of RecipeModel)(CurrentRecipe.Id))
            RequestNavigation()("/Views/RecipeEdit.xaml")
        End Sub

        ''' <summary>
        '''     Delete an existing recipe with confirmation
        ''' </summary>
        Private Sub _DeleteRecipe()
            If Not Dialog.ShowMessage("Confirm Delete", "Are you sure you wish to delete the recipe?", True) Then
                Return
            End If

            Dim currentRecipe = App.Database.Load(Of RecipeModel)(Me.CurrentRecipe.Id)

            For Each ingredient In currentRecipe.Ingredients
                App.Database.Delete(ingredient)
            Next ingredient
            App.Database.Delete(currentRecipe)
            Me.CurrentRecipe = Nothing

            ' refresh the recipe list
            RaisePropertyChanged(Function() Recipes)
        End Sub

        ''' <summary>
        '''     Can we edit?
        ''' </summary>
        ''' <returns>True if a recipe is selected</returns>
        Private Function _CanEditRecipe() As Boolean
            Return CurrentRecipe IsNot Nothing
        End Function

        ''' <summary>
        '''     Application buttons
        ''' </summary>
        Public Overrides ReadOnly Property ApplicationBindings() As IEnumerable(Of IActionCommand(Of Object))
            Get
                Return {AddRecipe, EditRecipe, DeleteRecipe}
            End Get
        End Property

        ''' <summary>
        '''     Tombstone
        ''' </summary>
        Public Sub Deactivate() Implements ITombstoneFriendly.Deactivate
            Dim tombstone = New TombstoneModel With {.SyncType = GetType(IMainViewModel)}
            If CurrentCategory IsNot Nothing Then
                tombstone.State.Add(ExtractPropertyName(Function() CurrentCategory), CurrentCategory.Id)
            End If
            If CurrentRecipe IsNot Nothing Then
                tombstone.State.Add(ExtractPropertyName(Function() CurrentRecipe), CurrentRecipe.Id)
            End If
            App.Database.Save(tombstone)
        End Sub

        ''' <summary>
        '''     Return from tombstone
        ''' </summary>
        Public Sub Activate() Implements ITombstoneFriendly.Activate
            Dim saved = App.Database.Load(Of TombstoneModel)(GetType(IMainViewModel))

            If saved Is Nothing Then
                Return
            End If

            Dim categoryId = saved.TryGet(ExtractPropertyName(Function() CurrentCategory), 0)

            If categoryId > 0 Then
                CurrentCategory = (
                    From c In Categories
                    Where c.Id = categoryId
                    Select c).FirstOrDefault()
            End If
            Dim currentRecipeId = saved.TryGet(ExtractPropertyName(Function() CurrentRecipe), 0)
            If currentRecipeId > 0 Then
                CurrentRecipe = (
                    From r In Recipes
                    Where r.Id = currentRecipeId
                    Select r).FirstOrDefault()
            End If
        End Sub
    End Class
End Namespace