﻿Imports System.Collections.Generic
Imports System.Collections.ObjectModel
Imports System.Linq
Imports SterlingRecipes.Contracts
Imports SterlingRecipes.Database
Imports SterlingRecipes.Messages
Imports SterlingRecipes.Models
Imports UltraLight.MVVM
Imports UltraLight.MVVM.Contracts

Namespace ViewModels
    ''' <summary>
    '''     Recipe view model
    ''' </summary>
    ''' <remarks>
    '''     Handles a text editor message and an ingredient message
    ''' </remarks>
    Public Class RecipeViewModel
        Inherits BaseViewModel
        Implements ITombstoneFriendly, IRecipeViewModel, IEventSink(Of ITextEditorViewModel), IEventSink(Of IngredientMessage)
        ' store the original recipe
        Private _recipe As RecipeModel

        ''' <summary>
        '''     Use these to flag when events are processed to avoid the tombstone recovery from overwriting
        '''     the message results
        ''' </summary>
        Private _instructionSink As Boolean

        Private _ingredientsSink As Boolean

        Private Const RECIPE_ID As String = "RecipeId"
        Private Const EDIT_RECIPE As String = "Edit Recipe"
        Private Const ADD_RECIPE As String = "Add Recipe"

        ''' <summary>
        '''     Set it all up and subscribe to events
        ''' </summary>
        Public Sub New()
            TextCommand = New ActionCommand(Of Object)(Sub(o)
                                                           UltraLightLocator.GetViewModel(Of ITextEditorViewModel)().BeginEdit("Instructions", Instructions)
                                                           RequestNavigation()("/Views/TextEditor.xaml")
                                                       End Sub)

            Ingredients = New ObservableCollection(Of IngredientModel)()

            AddIngredient = New ActionCommand(Of Object)(Sub(o) _AddIngredient())

            DeleteIngredient = New ActionCommand(Of IngredientModel)(AddressOf _DeleteIngredient, Function(i) i IsNot Nothing)

            SaveRecipe = New ActionCommand(Of Object)(Sub(o) _SaveRecipe())

            Cancel = New ActionCommand(Of Object)(Sub(o) _Cancel())

            UltraLightLocator.EventAggregator.Subscribe(Of ITextEditorViewModel)(Me)
            UltraLightLocator.EventAggregator.Subscribe(Of IngredientMessage)(Me)
        End Sub

        ''' <summary>
        '''     Name of the recipe
        ''' </summary>
        Private _recipeName As String
        Public Property RecipeName() As String
            Get
                Return _recipeName
            End Get
            Set(ByVal value As String)
                _recipeName = value
                RaisePropertyChanged(Function() RecipeName)
            End Set
        End Property

        Public ReadOnly Property RecipeNameAdapter() As String Implements IRecipeViewModel.RecipeName
            Get
                Return _recipeName
            End Get
        End Property

        ''' <summary>
        '''     Title (add/edit)
        ''' </summary>
        Private _title As String
        Public Property Title() As String
            Get
                Return _title
            End Get
            Set(ByVal value As String)
                _title = value
                RaisePropertyChanged(Function() Title)
            End Set
        End Property

        Public ReadOnly Property TitleAdapter() As String Implements IRecipeViewModel.Title
            Get
                Return _title
            End Get
        End Property

        Public Overrides Function CancelBackRequest() As Boolean
            If Not Dialog.ShowMessage("Confirm Cancel", "Are you sure you wish to cancel? All changes will be lost!", True) Then
                Return True
            End If

            App.Database.Delete(GetType(TombstoneModel), GetType(IRecipeViewModel))
            Return False
        End Function

        ''' <summary>
        '''     Category for the recipe
        ''' </summary>
        Private _recipeCategory As CategoryModel
        Public Property RecipeCategory() As CategoryModel
            Get
                Return _recipeCategory
            End Get
            Set(ByVal value As CategoryModel)
                _recipeCategory = value
                RaisePropertyChanged(Function() RecipeCategory)
            End Set
        End Property

        Public ReadOnly Property RecipeCategoryAdapter() As CategoryModel Implements IRecipeViewModel.RecipeCategory
            Get
                Return _recipeCategory
            End Get
        End Property

        ''' <summary>
        '''     Available categories
        ''' </summary>
        Public ReadOnly Property Categories() As IEnumerable(Of CategoryModel) Implements IRecipeViewModel.Categories
            Get
                Return UltraLightLocator.GetViewModel(Of IMainViewModel)().Categories
            End Get
        End Property

        ''' <summary>
        '''     Ingredients for this recipe
        ''' </summary>
        Private privateIngredients As ObservableCollection(Of IngredientModel)
        Public Property Ingredients() As ObservableCollection(Of IngredientModel)
            Get
                Return privateIngredients
            End Get
            Private Set(ByVal value As ObservableCollection(Of IngredientModel))
                privateIngredients = value
            End Set
        End Property

        Public ReadOnly Property IngredientsAdapter() As ObservableCollection(Of IngredientModel) Implements IRecipeViewModel.Ingredients
            Get
                Return privateIngredients
            End Get
        End Property

        ''' <summary>
        '''     Current ingredient to work with
        ''' </summary>
        Private _selectedIngredient As IngredientModel
        Public Property SelectedIngredient() As IngredientModel
            Get
                Return _selectedIngredient
            End Get
            Set(ByVal value As IngredientModel)
                _selectedIngredient = value
                RaisePropertyChanged(Function() SelectedIngredient)
                If value IsNot Nothing Then
                    _EditIngredient()
                End If
            End Set
        End Property

        Public ReadOnly Property SelectedIngredientAdapter() As IngredientModel Implements IRecipeViewModel.SelectedIngredient
            Get
                Return _selectedIngredient
            End Get
        End Property

        ''' <summary>
        '''     Instructions
        ''' </summary>
        Private _instructions As String
        Public Property Instructions() As String
            Get
                Return _instructions
            End Get
            Set(ByVal value As String)
                _instructions = value
                RaisePropertyChanged(Function() Instructions)
            End Set
        End Property

        Public ReadOnly Property InstructionsAdapter() As String Implements IRecipeViewModel.Instructions
            Get
                Return _instructions
            End Get
        End Property

        ''' <summary>
        '''     Edit the text (instructions)
        ''' </summary>
        Private privateTextCommand As IActionCommand(Of Object)
        Public Property TextCommand() As IActionCommand(Of Object)
            Get
                Return privateTextCommand
            End Get
            Private Set(ByVal value As IActionCommand(Of Object))
                privateTextCommand = value
            End Set
        End Property

        ''' <summary>
        '''     Add a new ingredient
        ''' </summary>
        Private privateAddIngredient As IActionCommand(Of Object)
        Public Property AddIngredient() As IActionCommand(Of Object)
            Get
                Return privateAddIngredient
            End Get
            Private Set(ByVal value As IActionCommand(Of Object))
                privateAddIngredient = value
            End Set
        End Property

        Public ReadOnly Property AddIngredientAdapter() As IActionCommand(Of Object) Implements IRecipeViewModel.AddIngredient
            Get
                Return privateAddIngredient
            End Get
        End Property

        ''' <summary>
        '''     Delete the existing ingredient
        ''' </summary>
        Private privateDeleteIngredient As IActionCommand(Of IngredientModel)
        Public Property DeleteIngredient() As IActionCommand(Of IngredientModel)
            Get
                Return privateDeleteIngredient
            End Get
            Private Set(ByVal value As IActionCommand(Of IngredientModel))
                privateDeleteIngredient = value
            End Set
        End Property

        Public ReadOnly Property DeleteIngredientAdapter() As IActionCommand(Of IngredientModel) Implements IRecipeViewModel.DeleteIngredient
            Get
                Return privateDeleteIngredient
            End Get
        End Property

        ''' <summary>
        '''     Save the recipe
        ''' </summary>
        Private privateSaveRecipe As IActionCommand(Of Object)
        Public Property SaveRecipe() As IActionCommand(Of Object)
            Get
                Return privateSaveRecipe
            End Get
            Private Set(ByVal value As IActionCommand(Of Object))
                privateSaveRecipe = value
            End Set
        End Property

        Public ReadOnly Property SaveRecipeAdapter() As IActionCommand(Of Object) Implements IRecipeViewModel.SaveRecipe
            Get
                Return privateSaveRecipe
            End Get
        End Property

        ''' <summary>
        '''     Cancel
        ''' </summary>
        Private privateCancel As IActionCommand(Of Object)
        Public Property Cancel() As IActionCommand(Of Object)
            Get
                Return privateCancel
            End Get
            Private Set(ByVal value As IActionCommand(Of Object))
                privateCancel = value
            End Set
        End Property

        Public ReadOnly Property CancelAdapter() As IActionCommand(Of Object) Implements IRecipeViewModel.Cancel
            Get
                Return privateCancel
            End Get
        End Property

        ''' <summary>
        '''     Bindings
        ''' </summary>
        Public Overrides ReadOnly Property ApplicationBindings() As IEnumerable(Of IActionCommand(Of Object))
            Get
                Return {AddIngredient, SaveRecipe, Cancel}
            End Get
        End Property

        ''' <summary>
        '''     Delete the ingredient
        ''' </summary>
        ''' <param name="ingredientModel">The ingredient to delete</param>
        Private Sub _DeleteIngredient(ByVal ingredientModel As IngredientModel)
            Ingredients.Remove(ingredientModel)
            RaisePropertyChanged(Function() Ingredients)
        End Sub

        ''' <summary>
        '''     Edit an ingredient
        ''' </summary>
        Private Sub _EditIngredient()
            UltraLightLocator.GetViewModel(Of IIngredientViewModel)().EditIngredient(SelectedIngredient)
            SelectedIngredient = Nothing
            RequestNavigation()("/Views/IngredientEdit.xaml")
        End Sub

        ''' <summary>
        '''     Add a new ingredient
        ''' </summary>
        Private Sub _AddIngredient()
            UltraLightLocator.GetViewModel(Of IIngredientViewModel)().AddIngredient()
            SelectedIngredient = Nothing
            RequestNavigation()("/Views/IngredientEdit.xaml")
        End Sub

        ''' <summary>
        '''     Save the recipe
        ''' </summary>
        Private Sub _SaveRecipe()
            Const CANNOT_SAVE As String = "Cannot Save"

            If String.IsNullOrEmpty(RecipeName) Then
                Dialog.ShowMessage(CANNOT_SAVE, "The recipe name is required.", False)
                Return
            End If

            If String.IsNullOrEmpty(Instructions) Then
                Dialog.ShowMessage(CANNOT_SAVE, "You must enter instructions.", False)
                Return
            End If

            If Ingredients.Count < 1 Then
                Dialog.ShowMessage(CANNOT_SAVE, "You need at least one ingredient.", False)
                Return
            End If

            _recipe.Name = RecipeName
            _recipe.Instructions = Instructions
            _recipe.Category = RecipeCategory

            ' set up ingredients on the new recipe
            If _recipe.Ingredients IsNot Nothing Then
                _recipe.Ingredients.Clear()
            Else
                _recipe.Ingredients = New List(Of IngredientModel)()
            End If

            ' purge any recipes that were removed
            If _recipe.Id > 0 Then
                Dim recipeOnDisk = App.Database.Load(Of RecipeModel)(_recipe.Id)
                For Each ingredient In
                    From _ingredient In recipeOnDisk.Ingredients
                    Let exists = (
                        From i In Ingredients
                        Where i.Id = _ingredient.Id
                        Select i).Any()
                    Where (Not exists)
                    Select _ingredient
                    App.Database.Delete(GetType(IngredientModel), ingredient.Id)
                Next ingredient
            Else
                ' generate an id
                App.Database.Save(_recipe)
            End If

            Dim order = 0

            ' iterate the ingredients
            For Each ingredient In Ingredients.OrderBy(Function(i) i.Order)
                ingredient.Recipe = _recipe
                order += 1
                ingredient.Order = order

                ' generate an id for the ingredient
                If ingredient.Id < 1 Then
                    App.Database.Save(ingredient)
                End If

                _recipe.Ingredients.Add(ingredient)
            Next ingredient

            App.Database.Save(_recipe)

            App.Database.Delete(GetType(TombstoneModel), GetType(IRecipeViewModel))
            App.Database.Flush() ' significant changes so let's make sure we get these on disk

            GoBack()()
        End Sub

        ''' <summary>
        '''     Cancel
        ''' </summary>
        Private Sub _Cancel()
            If Not Dialog.ShowMessage("Confirm Cancel", "Are you sure you wish to cancel? All changes will be lost!", True) Then
                Return
            End If

            App.Database.Delete(GetType(TombstoneModel), GetType(IRecipeViewModel))
            GoBack()()
        End Sub

        ''' <summary>
        '''     Edit the recipe
        ''' </summary>
        ''' <param name="recipe"></param>
        Public Sub EditRecipe(ByVal recipe As RecipeModel) Implements IRecipeViewModel.EditRecipe
            Title = EDIT_RECIPE
            _recipe = App.Database.Load(Of RecipeModel)(recipe.Id)
            RecipeName = _recipe.Name
            Instructions = _recipe.Instructions
            Ingredients.Clear()
            For Each ingredient In _recipe.Ingredients
                Ingredients.Add(ingredient)
            Next ingredient
            _selectedIngredient = Nothing
            RaisePropertyChanged(Function() SelectedIngredient)
            RaisePropertyChanged(Function() Categories)
            RecipeCategory = recipe.Category

            App.Database.Delete(GetType(TombstoneModel), GetType(IRecipeViewModel))
            _instructionSink = False
            _ingredientsSink = False
        End Sub

        ''' <summary>
        '''     Add a new recipe
        ''' </summary>
        Public Sub AddRecipe() Implements IRecipeViewModel.AddRecipe
            _recipe = New RecipeModel()
            Title = ADD_RECIPE
            RecipeName = String.Empty
            Instructions = String.Empty
            Ingredients.Clear()
            SelectedIngredient = Nothing
            RecipeCategory = If(UltraLightLocator.GetViewModel(Of IMainViewModel)().CurrentCategory, UltraLightLocator.GetViewModel(Of IMainViewModel)().Categories.First())
            RaisePropertyChanged(Function() Categories)
            App.Database.Delete(GetType(TombstoneModel), GetType(IRecipeViewModel))
            _instructionSink = False
            _ingredientsSink = False
        End Sub

        ''' <summary>
        '''     Handle the text edit event (update the instructions)
        ''' </summary>
        ''' <param name="publishedEvent">The view model handling the text</param>
        Public Sub HandleEvent(ByVal publishedEvent As ITextEditorViewModel) Implements IEventSink(Of SterlingRecipes.Contracts.ITextEditorViewModel).HandleEvent
            If publishedEvent Is Nothing Then
                Return
            End If

            If String.IsNullOrEmpty(publishedEvent.Text) Then
                Return
            End If

            Instructions = publishedEvent.Text
            _instructionSink = True
        End Sub

        ''' <summary>
        '''     Tombstone
        ''' </summary>
        Public Sub Deactivate() Implements ITombstoneFriendly.Deactivate
            If _recipe Is Nothing Then
                Return
            End If
            Dim tombstone = New TombstoneModel With {.SyncType = GetType(IRecipeViewModel)}
            tombstone.State.Add(RECIPE_ID, _recipe.Id)
            tombstone.State.Add(ExtractPropertyName(Function() RecipeCategory), RecipeCategory.Id)
            tombstone.State.Add(ExtractPropertyName(Function() RecipeName), RecipeName)
            tombstone.State.Add(ExtractPropertyName(Function() Title), Title)
            tombstone.State.Add(ExtractPropertyName(Function() Instructions), Instructions)

            Dim ingredients = Me.Ingredients.Where(Function(i) i IsNot Nothing).OrderBy(Function(i) i.Order).Select(Function(ingredient) New IngredientCacheModel With {.IngredientId = ingredient.Id, .Food = ingredient.Food, .Amount = ingredient.Amount}).ToList()
            tombstone.State.Add(ExtractPropertyName(Function() Me.Ingredients), ingredients)
            App.Database.Save(tombstone)
        End Sub

        ''' <summary>
        '''     Return from tombstone
        ''' </summary>
        Public Sub Activate() Implements ITombstoneFriendly.Activate
            Dim tombstone = App.Database.Load(Of TombstoneModel)(GetType(IRecipeViewModel))
            If tombstone Is Nothing Then
                Return
            End If
            Dim recipeId = tombstone.TryGet(RECIPE_ID, -1)
            _recipe = If(recipeId > 0, App.Database.Load(Of RecipeModel)(recipeId), New RecipeModel())

            Dim categoryId = tombstone.TryGet(ExtractPropertyName(Function() RecipeCategory), -1)
            If categoryId > 0 Then
                RecipeCategory = (
                    From c In UltraLightLocator.GetViewModel(Of IMainViewModel)().Categories
                    Where c.Id = categoryId
                    Select c).FirstOrDefault()
            End If

            RecipeName = tombstone.TryGet(ExtractPropertyName(Function() RecipeName), String.Empty)
            Title = tombstone.TryGet(ExtractPropertyName(Function() Title), String.Empty)

            If _instructionSink Then
                ' again, event aggregator sent message so don't overwrite with tombstone
                _instructionSink = False
            Else
                Instructions = tombstone.TryGet(ExtractPropertyName(Function() Instructions), String.Empty)
            End If

            If _ingredientsSink Then
                _ingredientsSink = False
            Else
                ' rebuild ingredients
                Ingredients.Clear()
                Dim order = 1
                For Each ingredientModel In
                    From ingredient In tombstone.TryGet(ExtractPropertyName(Function() Ingredients), Enumerable.Empty(Of IngredientCacheModel)())
                    Select New IngredientModel With {.Id = ingredient.IngredientId, .Order = order + 1, .Food = ingredient.Food, .Amount = ingredient.Amount}
                    Ingredients.Add(ingredientModel)
                Next ingredientModel
            End If
        End Sub

        ''' <summary>
        '''     Handle an ingredient event (update/remove)
        ''' </summary>
        ''' <param name="publishedEvent"></param>
        Public Sub HandleEvent(ByVal publishedEvent As IngredientMessage) Implements IEventSink(Of SterlingRecipes.Messages.IngredientMessage).HandleEvent
            If publishedEvent.Action.Equals(IngredientAction.Save) Then
                SelectedIngredient = Nothing
                Dim existingIngredient = (
                    From i In Ingredients
                    Where i.Id = publishedEvent.Model.Id
                    Select i).FirstOrDefault()
                If existingIngredient IsNot Nothing Then
                    Ingredients.Remove(existingIngredient)
                End If
                Ingredients.Add(publishedEvent.Model)
                _ingredientsSink = True
            Else
                If DeleteIngredient.CanExecute(publishedEvent.Model) Then
                    DeleteIngredient.Execute(publishedEvent.Model)
                End If
            End If
        End Sub
    End Class
End Namespace