﻿Imports System
Imports System.ComponentModel
Imports System.Linq.Expressions
Imports System.Reflection

''' <summary>
'''     Base class for items that implement property changed
''' </summary>
''' <remarks>
'''     Code for this section was inspired by two other projects: 
'''     1. The PRISM implementation and MVVM guidance here: http://compositewpf.codeplex.com/
'''     2. The Caliburn.Micro framework here: http://caliburnmicro.codeplex.com/ 
''' </remarks>
Public MustInherit Class BaseNotify
    Implements INotifyPropertyChanged
    ''' <summary>
    ''' Raised when a property on this object has a new value.
    ''' </summary>
    Public Event PropertyChanged As PropertyChangedEventHandler Implements INotifyPropertyChanged.PropertyChanged

    ''' <summary>
    ''' Raises this object's PropertyChanged event.
    ''' </summary>
    ''' <param name="propertyName">The property that has a new value.</param>
    Protected Overridable Sub RaisePropertyChanged(ByVal propertyName As String)
        Dim handler = PropertyChangedEvent
        If handler IsNot Nothing Then
            handler(Me, New PropertyChangedEventArgs(propertyName))
        End If
    End Sub

    ''' <summary>
    ''' Raises this object's PropertyChanged event for each of the properties.
    ''' </summary>
    ''' <param name="propertyNames">The properties that have a new value.</param>
    Protected Sub RaisePropertyChanged(ByVal ParamArray propertyNames() As String)
        For Each name In propertyNames
            RaisePropertyChanged(name)
        Next name
    End Sub

    ''' <summary>
    ''' Raises this object's PropertyChanged event.
    ''' </summary>
    ''' <typeparam name="T">The type of the property that has a new value</typeparam>
    ''' <param name="propertyExpression">A Lambda expression representing the property that has a new value.</param>
    Protected Sub RaisePropertyChanged(Of T)(ByVal propertyExpression As Expression(Of Func(Of T)))
        Dim propertyName = ExtractPropertyName(propertyExpression)
        RaisePropertyChanged(propertyName)
    End Sub

    ''' <summary>
    '''     Takes an expression and pulls out the property name
    ''' </summary>
    ''' <typeparam name="T">Type of the property</typeparam>
    ''' <param name="propertyExpression">An expression that returns the property</param>
    ''' <returns>The name of the property</returns>
    Protected Function ExtractPropertyName(Of T)(ByVal propertyExpression As Expression(Of Func(Of T))) As String
        If propertyExpression Is Nothing Then
            Throw New ArgumentNullException("propertyExpression")
        End If

        Dim memberExpression = TryCast(propertyExpression.Body, MemberExpression)
        If memberExpression Is Nothing Then
            Throw New ArgumentException("propertyExpression")
        End If

        Dim [property] = TryCast(memberExpression.Member, PropertyInfo)
        If [property] Is Nothing Then
            Throw New ArgumentException("propertyExpression")
        End If

        Dim getMethod = [property].GetGetMethod(True)

        If getMethod Is Nothing Then
            ' this shouldn't happen - the expression would reject the property before reaching this far
            Throw New ArgumentException("propertyExpression")
        End If

        If getMethod.IsStatic Then
            Throw New ArgumentException("propertyExpression")
        End If

        Return memberExpression.Member.Name
    End Function
End Class