﻿Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Threading
Imports UltraLight.MVVM.Contracts

''' <summary>
'''     Implementation of event aggregator pattern using reactive extensions
''' </summary>
Public Class EventAggregatorService
    Implements IEventAggregator
    ''' <summary>
    '''     Mutex for locking
    ''' </summary>
    Private Shared ReadOnly _mutex As New Object()

    ''' <summary>
    '''     Subscribers
    ''' </summary>
    Private Shared ReadOnly _subscribers As New Dictionary(Of Type, List(Of EventEntry))()

    ''' <summary>
    '''     Set up the event for the first time
    ''' </summary>
    ''' <typeparam name="TEvent">The type of the event</typeparam>
    Private Shared Sub _RegisterEvent(Of TEvent)()
        If _subscribers.ContainsKey(GetType(TEvent)) Then
            Return
        End If

        Monitor.Enter(_mutex)
        If Not _subscribers.ContainsKey(GetType(TEvent)) Then
            _subscribers.Add(GetType(TEvent), New List(Of EventEntry)())
        End If
        Monitor.Exit(_mutex)
    End Sub

    ''' <summary>
    '''     Publish an event
    ''' </summary>
    ''' <typeparam name="TEvent">The type of the event to publish</typeparam>
    ''' <param name="sampleEvent">The event data</param>
    Public Sub Publish(Of TEvent)(ByVal sampleEvent As TEvent) Implements IEventAggregator.Publish
        If Not _subscribers.ContainsKey(GetType(TEvent)) Then
            ' no one is listening
            Return
        End If

        ' snapshot the list
        Monitor.Enter(_mutex)
        Dim subscribers = (
            From _sub In _subscribers(GetType(TEvent))
            Select _sub).ToArray()
        Monitor.Exit(_mutex)

        ' now filter through and mark any dead subscriptions
        Dim dead = New List(Of EventEntry)()
        For Each [sub] In subscribers
            Dim sink = TryCast([sub].Subscriber.Target, IEventSink(Of TEvent))
            If sink Is Nothing OrElse (Not [sub].Subscriber.IsAlive) Then
                dead.Add([sub])
            Else
                If [sub].ExecuteOnDispatcher Then
                    UltraLightLocator.ExecuteOnUI(Sub() sink.HandleEvent(sampleEvent))
                Else
                    sink.HandleEvent(sampleEvent)
                End If
            End If
        Next [sub]

        ' scrub the dead subscriptions
        Monitor.Enter(_mutex)
        For Each deadSub In dead.Where(Function(_deadSub) _subscribers(GetType(TEvent)).Contains(_deadSub))
            _subscribers(GetType(TEvent)).Remove(deadSub)
        Next deadSub
        Monitor.Exit(_mutex)
    End Sub

    ''' <summary>
    '''     Get a handle to subscribe/listen to an event
    ''' </summary>
    ''' <typeparam name="TEvent">The event type to listen for</typeparam>
    ''' <param name="subscriber">The subscriber</param>
    Public Sub Subscribe(Of TEvent)(ByVal subscriber As IEventSink(Of TEvent)) Implements IEventAggregator.Subscribe
        _RegisterEvent(Of TEvent)()
        _subscribers(GetType(TEvent)).Add(EventEntry.Create(False, New WeakReference(subscriber)))
    End Sub

    ''' <summary>
    '''     Get a handle to subscribe/listen to an event
    ''' </summary>
    ''' <typeparam name="TEvent">The event type to listen for</typeparam>
    ''' <param name="subscriber">The subscriber</param>
    Public Sub SubscribeOnDispatcher(Of TEvent)(ByVal subscriber As IEventSink(Of TEvent)) Implements IEventAggregator.SubscribeOnDispatcher
        _RegisterEvent(Of TEvent)()
        _subscribers(GetType(TEvent)).Add(EventEntry.Create(True, New WeakReference(subscriber)))
    End Sub

    ''' <summary>
    '''     Unsubscribe
    ''' </summary>
    ''' <typeparam name="TEvent"></typeparam>
    ''' <param name="unsubscriber"></param>
    Public Sub Unsubscribe(Of TEvent)(ByVal unsubscriber As IEventSink(Of TEvent)) Implements IEventAggregator.Unsubscribe
        If Not _subscribers.ContainsKey(GetType(TEvent)) Then
            Return
        End If
        Dim unsub = (
            From s In _subscribers(GetType(TEvent))
            Where ReferenceEquals(s.Subscriber.Target, unsubscriber)
            Select s).ToList()
        Monitor.Enter(_mutex)
        For Each subCast In unsub.Select(Function(_sub) _sub).Where(Function(_subCast) _subscribers(GetType(TEvent)).Contains(_subCast))
            _subscribers(GetType(TEvent)).Remove(subCast)
        Next subCast
        Monitor.Exit(_mutex)
		End Sub
End Class