﻿Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Windows
Imports System.Windows.Controls
Imports System.Windows.Media
Imports System.Windows.Threading
Imports Microsoft.Phone.Controls
Imports Microsoft.Phone.Shell
Imports UltraLight.MVVM.Contracts

''' <summary>
'''     Global MVVM Manager
''' </summary>
Public Module UltraLightLocator
    ''' <summary>
    '''     Hold View model references
    ''' </summary>
    Private ReadOnly _viewModels As New Dictionary(Of Type, IViewModel)()

    Private _eventAggregator As IEventAggregator

    Public ReadOnly Property EventAggregator() As IEventAggregator
        Get
            If _eventAggregator IsNot Nothing Then
                Return _eventAggregator
            Else
                _eventAggregator = New EventAggregatorService()
                Return _eventAggregator
            End If
        End Get
    End Property

    ''' <summary>
    '''     When a page is tombstoned
    ''' </summary>
    ''' <param name="phonePage">The page</param>
    <System.Runtime.CompilerServices.Extension()> _
    Public Sub DeactivatePage(Of T As IViewModel)(ByVal phonePage As PhoneApplicationPage)
        Dim viewModel = GetViewModel(Of T)()
        If TypeOf viewModel Is ITombstoneFriendly Then
            CType(viewModel, ITombstoneFriendly).Deactivate()
        End If
    End Sub

    ''' <summary>
    '''     Activate page
    ''' </summary>
    ''' <remarks>
    '''     We allow a control, not a page here, in case you have a Panorama or Pivot control,
    '''     you'll want to hook to that loaded event instead of the page so it initializes
    '''     correctly
    ''' </remarks>
    ''' <param name="element">The root element - page, Grid, Pivot, etc.</param>
    <System.Runtime.CompilerServices.Extension()> _
    Public Sub ActivatePage(Of T As IViewModel)(ByVal element As FrameworkElement)
        Dim viewModel = GetViewModel(Of T)()
        Dim loaded As RoutedEventHandler = Nothing
        loaded = Sub(o, e)
                     ' ReSharper disable AccessToModifiedClosure
                     RemoveHandler CType(o, FrameworkElement).Loaded, loaded
                     ' ReSharper restore AccessToModifiedClosure
                     If TypeOf viewModel Is ITombstoneFriendly Then
                         ' we queue this on the dispatcher even if we know we're already on the UI thread
                         ' this allows controls like Pivot to finish their initialization before we activate 
                         ' the view model bindings
                         Deployment.Current.Dispatcher.BeginInvoke(Sub() CType(viewModel, ITombstoneFriendly).Activate())
                     End If
                 End Sub
        AddHandler element.Loaded, loaded
    End Sub

    ''' <summary>
    '''     Bind a page to a view model
    ''' </summary>
    ''' <param name="phonePage">The page</param>
    ''' <param name="root">The root element</param>
    <System.Runtime.CompilerServices.Extension()> _
    Public Sub BindToViewModel(Of T As IViewModel)(ByVal phonePage As PhoneApplicationPage, ByVal root As FrameworkElement)
        If DesignerProperties.IsInDesignTool Then
            Return
        End If

        Dim loaded As RoutedEventHandler = Nothing

        Dim vm = GetViewModel(Of T)() ' access to modified closure

        loaded = Sub(o, e)
                     Dim page = TryCast(o, PhoneApplicationPage)
                     If page Is Nothing Then
                         Return
                     End If

                     ' unhook loaded so we do this only once
                     ' ReSharper disable AccessToModifiedClosure
                     RemoveHandler page.Loaded, loaded
                     ' ReSharper restore AccessToModifiedClosure

                     ' provide an implementation for the dialog
                     vm.Dialog = New Dialog()

                     vm.GoToVisualState = Sub(state, useTransitions) ExecuteOnUI(Function() VisualStateManager.GoToState(phonePage, state, useTransitions))


                     ' provide an implementation for navigation
                     vm.RequestNavigation = Function(newPage) page.NavigationService.Navigate(New Uri(newPage, UriKind.Relative))

                     vm.GoBack = AddressOf page.NavigationService.GoBack

                     AddHandler page.BackKeyPress, Sub(sender, args) args.Cancel = vm.CancelBackRequest()

                     ' bind the view model
                     root.DataContext = vm

                     ' bind the application bar buttons
                     Dim idx = 0

                     If page.ApplicationBar Is Nothing OrElse page.ApplicationBar.Buttons Is Nothing Then
                         Return
                     End If

                     For Each command In vm.ApplicationBindings
                         If page.ApplicationBar.Buttons.Count <= idx Then
                             Exit For
                         End If

                         Dim appButton = TryCast(page.ApplicationBar.Buttons(idx), ApplicationBarIconButton)
                         idx += 1

                         If appButton Is Nothing Then
                             Continue For
                         End If
                         Dim command1 = command
                         AddHandler command.CanExecuteChanged, Sub(cmd, arg) appButton.IsEnabled = command1.CanExecute(Nothing)
                         AddHandler appButton.Click, Sub(cmd, arg) command1.Execute(Nothing)
                     Next command
                 End Sub

        AddHandler phonePage.Loaded, loaded
    End Sub

    ''' <summary>
    '''     Bind a user control to a view model
    ''' </summary>
    ''' <param name="userControl">The user control</param>
    ''' <param name="root">The root layout</param>
    <System.Runtime.CompilerServices.Extension()> _
    Public Sub BindToViewModel(Of T As IViewModel)(ByVal userControl As UserControl, ByVal root As FrameworkElement)
        If DesignerProperties.IsInDesignTool Then
            Return
        End If

        Dim loaded As RoutedEventHandler = Nothing

        Dim vm = GetViewModel(Of T)() ' access to modified closure

        loaded = Sub(o, e)
                     Dim control = TryCast(o, UserControl)
                     If control Is Nothing Then
                         Return
                     End If

                     ' unhook loaded so we do this only once
                     ' ReSharper disable AccessToModifiedClosure
                     RemoveHandler control.Loaded, loaded
                     ' ReSharper restore AccessToModifiedClosure

                     ' provide an implementation for the dialog
                     vm.Dialog = New Dialog()

                     vm.GoToVisualState = Sub(state, useTransitions) ExecuteOnUI(Function() VisualStateManager.GoToState(userControl, state, useTransitions))

                     ' hook to the navigation of the parent
                     Dim parent = VisualTreeHelper.GetParent(control)
                     Do While parent IsNot Nothing
                         If TypeOf parent Is PhoneApplicationPage Then
                             vm.RequestNavigation = Function(newPage) (CType(parent, PhoneApplicationPage)).NavigationService.Navigate(New Uri(newPage, UriKind.Relative))
                             vm.GoBack = AddressOf CType(parent, PhoneApplicationPage).NavigationService.GoBack
                             AddHandler CType(parent, PhoneApplicationPage).BackKeyPress, Sub(sender, args) args.Cancel = vm.CancelBackRequest()

                             Exit Do
                         End If
                         parent = VisualTreeHelper.GetParent(parent)
                     Loop

                     ' bind the view model
                     root.DataContext = vm
                 End Sub

        AddHandler userControl.Loaded, loaded
    End Sub

    ''' <summary>
    '''     Register a view model by type
    ''' </summary>
    ''' <typeparam name="T">The type, preferably an interface for the view model</typeparam>
    ''' <param name="viewModel">The instance to register</param>
    Public Sub Register(Of T As IViewModel)(ByVal viewModel As T)
        _viewModels.Add(GetType(T), viewModel)
    End Sub

    ''' <summary>
    '''     Get the view model for the type
    ''' </summary>
    ''' <typeparam name="T">The type, preferably an interface for the view model</typeparam>
    ''' <returns>The instance of the view model previously registered</returns>
    Public Function GetViewModel(Of T As IViewModel)() As T
        Return If(_viewModels.ContainsKey(GetType(T)), CType(_viewModels(GetType(T)), T), Nothing)
    End Function

    Private ReadOnly _dispatcher As Dispatcher = Deployment.Current.Dispatcher

    ''' <summary>
    '''     Execute on the UI thread
    ''' </summary>
    ''' <param name="action">Action to execute</param>
    Public Sub ExecuteOnUI(ByVal action As Action)
        If _dispatcher.CheckAccess() Then
            action()
        Else
            _dispatcher.BeginInvoke(action)
        End If
    End Sub
End Module